@@ -45,6 +45,8 @@
 
 #include "logging.h"
 #include "textfile.h"
+#include "../src/adapter.h"
+#include "../src/device.h"
 
 #include "error.h"
 #include "ipc.h"
@@ -56,6 +58,7 @@
 #include "headset.h"
 #include "gateway.h"
 #include "sink.h"
+#include "source.h"
 
 #define AUDIO_INTERFACE "org.bluez.Audio"
 
@@ -69,12 +72,18 @@ typedef enum {
 	AUDIO_STATE_CONNECTED,
 } audio_state_t;
 
+struct service_auth {
+	service_auth_cb cb;
+	void *user_data;
+};
+
 struct dev_priv {
 	audio_state_t state;
 
 	headset_state_t hs_state;
 	sink_state_t sink_state;
 	avctp_state_t avctp_state;
+	GSList *auths;
 
 	DBusMessage *conn_req;
 	DBusMessage *dc_req;
@@ -96,6 +105,8 @@ static void device_free(struct audio_device *dev)
 	if (dev->conn)
 		dbus_connection_unref(dev->conn);
 
+	btd_device_unref(dev->btd_dev);
+
 	if (priv) {
 		if (priv->control_timer)
 			g_source_remove(priv->control_timer);
@@ -254,11 +265,19 @@ static void device_remove_avdtp_timer(struct audio_device *dev)
 static gboolean headset_connect_timeout(gpointer user_data)
 {
 	struct audio_device *dev = user_data;
+	struct dev_priv *priv = dev->priv;
 
 	dev->priv->headset_timer = 0;
 
-	if (dev->headset)
-		headset_config_stream(dev, FALSE, NULL, NULL);
+	if (dev->headset == NULL)
+		return FALSE;
+
+	if (headset_config_stream(dev, FALSE, NULL, NULL) == 0) {
+		if (priv->state != AUDIO_STATE_CONNECTED &&
+				(priv->sink_state == SINK_STATE_CONNECTED ||
+				priv->sink_state == SINK_STATE_PLAYING))
+			device_set_state(dev, AUDIO_STATE_CONNECTED);
+	}
 
 	return FALSE;
 }
@@ -322,9 +341,16 @@ static void device_sink_cb(struct audio_device *dev,
 		}
 		if (priv->hs_state == HEADSET_STATE_DISCONNECTED)
 			device_set_state(dev, AUDIO_STATE_DISCONNECTED);
-		else if (old_state == SINK_STATE_CONNECTING &&
-				priv->hs_state == HEADSET_STATE_CONNECTED)
-			device_set_state(dev, AUDIO_STATE_CONNECTED);
+		else if (old_state == SINK_STATE_CONNECTING) {
+			switch (priv->hs_state) {
+			case HEADSET_STATE_CONNECTED:
+			case HEADSET_STATE_PLAY_IN_PROGRESS:
+			case HEADSET_STATE_PLAYING:
+				device_set_state(dev, AUDIO_STATE_CONNECTED);
+			default:
+				break;
+			}
+		}
 		break;
 	case SINK_STATE_CONNECTING:
 		device_remove_avdtp_timer(dev);
@@ -337,11 +363,14 @@ static void device_sink_cb(struct audio_device *dev,
 		if (dev->auto_connect) {
 			if (!dev->headset)
 				device_set_state(dev, AUDIO_STATE_CONNECTED);
-			if (priv->hs_state == HEADSET_STATE_DISCONNECTED)
+			else if (priv->hs_state == HEADSET_STATE_DISCONNECTED)
 				device_set_headset_timer(dev);
-			else if (priv->hs_state == HEADSET_STATE_CONNECTED)
+			else if (priv->hs_state == HEADSET_STATE_CONNECTED ||
+					priv->hs_state == HEADSET_STATE_PLAY_IN_PROGRESS ||
+					priv->hs_state == HEADSET_STATE_PLAYING)
 				device_set_state(dev, AUDIO_STATE_CONNECTED);
-		} else if (priv->hs_state != HEADSET_STATE_CONNECTED)
+		} else if (priv->hs_state == HEADSET_STATE_DISCONNECTED ||
+				priv->hs_state == HEADSET_STATE_CONNECTING)
 			device_set_state(dev, AUDIO_STATE_CONNECTED);
 		break;
 	case SINK_STATE_PLAYING:
@@ -390,28 +419,33 @@ static void device_headset_cb(struct audio_device *dev,
 			sink_shutdown(dev->sink);
 			break;
 		}
-		if (priv->sink_state == AVDTP_SESSION_STATE_DISCONNECTED)
+		if (priv->sink_state == SINK_STATE_DISCONNECTED)
 			device_set_state(dev, AUDIO_STATE_DISCONNECTED);
-		else if (old_state == HEADSET_STATE_CONNECT_IN_PROGRESS &&
-				priv->sink_state == SINK_STATE_CONNECTED)
+		else if (old_state == HEADSET_STATE_CONNECTING &&
+				(priv->sink_state == SINK_STATE_CONNECTED ||
+				priv->sink_state == SINK_STATE_PLAYING))
 			device_set_state(dev, AUDIO_STATE_CONNECTED);
 		break;
-	case HEADSET_STATE_CONNECT_IN_PROGRESS:
+	case HEADSET_STATE_CONNECTING:
 		device_remove_headset_timer(dev);
 		if (priv->sink_state == SINK_STATE_DISCONNECTED)
 			device_set_state(dev, AUDIO_STATE_CONNECTING);
 		break;
 	case HEADSET_STATE_CONNECTED:
-		if (old_state == HEADSET_STATE_PLAYING)
+		if (old_state == HEADSET_STATE_CONNECTED ||
+				old_state == HEADSET_STATE_PLAY_IN_PROGRESS ||
+				old_state == HEADSET_STATE_PLAYING)
 			break;
 		if (dev->auto_connect) {
 			if (!dev->sink)
 				device_set_state(dev, AUDIO_STATE_CONNECTED);
 			else if (priv->sink_state == SINK_STATE_DISCONNECTED)
 				device_set_avdtp_timer(dev);
-			else if (priv->sink_state == SINK_STATE_CONNECTED)
+			else if (priv->sink_state == SINK_STATE_CONNECTED ||
+					priv->sink_state == SINK_STATE_PLAYING)
 				device_set_state(dev, AUDIO_STATE_CONNECTED);
-		} else if (priv->sink_state != SINK_STATE_CONNECTED)
+		} else if (priv->sink_state == SINK_STATE_DISCONNECTED ||
+				priv->sink_state == SINK_STATE_CONNECTING)
 			device_set_state(dev, AUDIO_STATE_CONNECTED);
 		break;
 	case HEADSET_STATE_PLAY_IN_PROGRESS:
@@ -439,7 +473,8 @@ static DBusMessage *dev_connect(DBusConnection *conn, DBusMessage *msg,
 
 	if (dev->headset)
 		headset_config_stream(dev, FALSE, NULL, NULL);
-	else if (dev->sink) {
+
+	if (priv->state != AUDIO_STATE_CONNECTING && dev->sink) {
 		struct avdtp *session = avdtp_get(&dev->src, &dev->dst);
 
 		if (!session)
@@ -451,7 +486,7 @@ static DBusMessage *dev_connect(DBusConnection *conn, DBusMessage *msg,
 		avdtp_unref(session);
 	}
 
-	/* The previous call should cause a call to the state callback to
+	/* The previous calls should cause a call to the state callback to
 	 * indicate AUDIO_STATE_CONNECTING */
 	if (priv->state != AUDIO_STATE_CONNECTING)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
@@ -479,7 +514,7 @@ static DBusMessage *dev_disconnect(DBusConnection *conn, DBusMessage *msg,
 	priv->dc_req = dbus_message_ref(msg);
 
 	if (priv->hs_state != HEADSET_STATE_DISCONNECTED)
-		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
+		headset_shutdown(dev);
 	else if (dev->sink && priv->sink_state != SINK_STATE_DISCONNECTED)
 		sink_shutdown(dev->sink);
 	else {
@@ -535,6 +570,7 @@ static GDBusSignalTable dev_signals[] = {
 };
 
 struct audio_device *audio_device_register(DBusConnection *conn,
+					struct btd_device *device,
 					const char *path, const bdaddr_t *src,
 					const bdaddr_t *dst)
 {
@@ -545,6 +581,7 @@ struct audio_device *audio_device_register(DBusConnection *conn,
 
 	dev = g_new0(struct audio_device, 1);
 
+	dev->btd_dev = btd_device_ref(device);
 	dev->path = g_strdup(path);
 	bacpy(&dev->dst, dst);
 	bacpy(&dev->src, src);
@@ -612,19 +649,65 @@ gboolean audio_device_is_connected(struct audio_device *dev,
 
 void audio_device_unregister(struct audio_device *device)
 {
+	unix_device_removed(device);
+
 	if (device->headset)
 		headset_unregister(device);
 
 	if (device->sink)
 		sink_unregister(device);
 
+	if (device->source)
+		source_unregister(device);
+
 	if (device->control)
 		control_unregister(device);
 
-	unix_device_removed(device);
-
 	g_dbus_unregister_interface(device->conn, device->path,
 						AUDIO_INTERFACE);
 
 	device_free(device);
 }
+
+static void auth_cb(DBusError *derr, void *user_data)
+{
+	struct audio_device *dev = user_data;
+	struct dev_priv *priv = dev->priv;
+
+	while (priv->auths) {
+		struct service_auth *auth = priv->auths->data;
+
+		auth->cb(derr, auth->user_data);
+		priv->auths = g_slist_remove(priv->auths, auth);
+		g_free(auth);
+	}
+}
+
+int audio_device_request_authorization(struct audio_device *dev,
+					const char *uuid, service_auth_cb cb,
+					void *user_data)
+{
+	struct dev_priv *priv = dev->priv;
+	struct service_auth *auth;
+	int err;
+
+	auth = g_try_new0(struct service_auth, 1);
+	if (!auth)
+		return -ENOMEM;
+
+	auth->cb = cb;
+	auth->user_data = user_data;
+
+	priv->auths = g_slist_append(priv->auths, auth);
+	if (g_slist_length(priv->auths) > 1)
+		return 0;
+
+	err = btd_request_authorization(&dev->src, &dev->dst, uuid, auth_cb,
+					dev);
+	if (err < 0) {
+		priv->auths = g_slist_remove(priv->auths, auth);
+		g_free(auth);
+	}
+
+	return err;
+}
