@@ -125,6 +125,7 @@ struct pending_connect {
 	int err;
 	headset_state_t target_state;
 	GSList *callbacks;
+	uint16_t svclass;
 };
 
 struct headset {
@@ -153,6 +154,7 @@ struct headset {
 	gboolean cme_enabled;
 	gboolean cwa_enabled;
 	gboolean pending_ring;
+	gboolean inband_ring;
 	gboolean nrec;
 	gboolean nrec_req;
 
@@ -276,7 +278,7 @@ static const char *state2str(headset_state_t state)
 	switch (state) {
 	case HEADSET_STATE_DISCONNECTED:
 		return "disconnected";
-	case HEADSET_STATE_CONNECT_IN_PROGRESS:
+	case HEADSET_STATE_CONNECTING:
 		return "connecting";
 	case HEADSET_STATE_CONNECTED:
 	case HEADSET_STATE_PLAY_IN_PROGRESS:
@@ -441,6 +443,9 @@ static void pending_connect_finalize(struct audio_device *dev)
 	if (p == NULL)
 		return;
 
+	if (p->svclass)
+		bt_cancel_discovery(&dev->src, &dev->dst);
+
 	g_slist_foreach(p->callbacks, (GFunc) pending_connect_complete, dev);
 
 	g_slist_foreach(p->callbacks, (GFunc) g_free, NULL);
@@ -703,7 +708,7 @@ int telephony_event_reporting_rsp(void *telephony_device, cme_error_t err)
 	if (ret < 0)
 		return ret;
 
-	if (hs->state != HEADSET_STATE_CONNECT_IN_PROGRESS)
+	if (hs->state != HEADSET_STATE_CONNECTING)
 		return 0;
 
 	if (hs->hf_features & HF_FEATURE_CALL_WAITING_AND_3WAY &&
@@ -770,7 +775,7 @@ static int call_hold(struct audio_device *dev, const char *buf)
 	if (err < 0)
 		return err;
 
-	if (hs->state != HEADSET_STATE_CONNECT_IN_PROGRESS)
+	if (hs->state != HEADSET_STATE_CONNECTING)
 		return 0;
 
 	hfp_slc_complete(dev);
@@ -1169,6 +1174,8 @@ static void close_sco(struct audio_device *device)
 	struct headset *hs = device->headset;
 
 	if (hs->sco) {
+		int sock = g_io_channel_unix_get_fd(hs->sco);
+		shutdown(sock, SHUT_RDWR);
 		g_io_channel_shutdown(hs->sco, TRUE, NULL);
 		g_io_channel_unref(hs->sco);
 		hs->sco = NULL;
@@ -1292,11 +1299,8 @@ void headset_connect_cb(GIOChannel *chan, GError *err, gpointer user_data)
 		goto failed;
 	}
 
-	if (hs->tmp_rfcomm) {
-		hs->rfcomm = hs->tmp_rfcomm;
-		hs->tmp_rfcomm = NULL;
-	} else
-		hs->rfcomm = g_io_channel_ref(chan);
+	hs->rfcomm = hs->tmp_rfcomm;
+	hs->tmp_rfcomm = NULL;
 
 	ba2str(&dev->dst, hs_address);
 
@@ -1342,7 +1346,7 @@ failed:
 		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
 }
 
-static void headset_set_channel(struct headset *headset,
+static int headset_set_channel(struct headset *headset,
 				const sdp_record_t *record, uint16_t svc)
 {
 	int ch;
@@ -1350,7 +1354,7 @@ static void headset_set_channel(struct headset *headset,
 
 	if (sdp_get_access_protos(record, &protos) < 0) {
 		error("Unable to get access protos from headset record");
-		return;
+		return -1;
 	}
 
 	ch = sdp_get_proto_port(protos, RFCOMM_UUID);
@@ -1358,13 +1362,22 @@ static void headset_set_channel(struct headset *headset,
 	sdp_list_foreach(protos, (sdp_list_func_t) sdp_list_free, NULL);
 	sdp_list_free(protos, NULL);
 
-	if (ch > 0) {
-		headset->rfcomm_ch = ch;
-		debug("Discovered %s service on RFCOMM channel %d",
-			svc == HEADSET_SVCLASS_ID ? "Headset" : "Handsfree",
-			ch);
-	} else
+	if (ch <= 0) {
 		error("Unable to get RFCOMM channel from Headset record");
+		return -1;
+	}
+
+	headset->rfcomm_ch = ch;
+
+	if (svc == HANDSFREE_SVCLASS_ID) {
+		headset->hfp_handle = record->handle;
+		debug("Discovered Handsfree service on channel %d", ch);
+	} else {
+		headset->hsp_handle = record->handle;
+		debug("Discovered Headset service on channel %d", ch);
+	}
+
+	return 0;
 }
 
 static void get_record_cb(sdp_list_t *recs, int err, gpointer user_data)
@@ -1373,13 +1386,17 @@ static void get_record_cb(sdp_list_t *recs, int err, gpointer user_data)
 	struct headset *hs = dev->headset;
 	struct pending_connect *p = hs->pending;
 	sdp_record_t *record = NULL;
-	sdp_list_t *classes = NULL;
+	sdp_list_t *r;
 	uuid_t uuid;
 
+	assert(hs->pending != NULL);
+
 	if (err < 0) {
-		error("Unable to get service record: %s (%d)", strerror(-err),
-			-err);
-		goto failed_not_supported;
+		error("Unable to get service record: %s (%d)",
+							strerror(-err), -err);
+		p->err = -err;
+		error_connection_attempt_failed(dev->conn, p->msg, p->err);
+		goto failed;
 	}
 
 	if (!recs || !recs->data) {
@@ -1387,41 +1404,42 @@ static void get_record_cb(sdp_list_t *recs, int err, gpointer user_data)
 		goto failed_not_supported;
 	}
 
-	record = recs->data;
+	sdp_uuid16_create(&uuid, p->svclass);
 
-	if (sdp_get_service_classes(record, &classes) < 0) {
-		error("Unable to get service classes from record");
+	for (r = recs; r != NULL; r = r->next) {
+		sdp_list_t *classes;
+		uuid_t class;
+
+		record = r->data;
+
+		if (sdp_get_service_classes(record, &classes) < 0) {
+			error("Unable to get service classes from record");
+			continue;
+		}
+
+		memcpy(&class, classes->data, sizeof(uuid));
+
+		sdp_list_free(classes, free);
+
+
+		if (sdp_uuid_cmp(&class, &uuid) == 0)
+			break;
+	}
+
+	if (r == NULL) {
+		error("No record found with UUID 0x%04x", p->svclass);
 		goto failed_not_supported;
 	}
 
-	memcpy(&uuid, classes->data, sizeof(uuid));
-
-	if (!sdp_uuid128_to_uuid(&uuid) || uuid.type != SDP_UUID16) {
-		error("Not a 16 bit UUID");
-		goto failed_not_supported;
-	}
-
-	if (hs->search_hfp) {
-		if (uuid.value.uuid16 != HANDSFREE_SVCLASS_ID) {
-			error("Service record didn't contain the HFP UUID");
-			goto failed_not_supported;
-		}
-		hs->hfp_handle = record->handle;
-	} else {
-		if (uuid.value.uuid16 != HEADSET_SVCLASS_ID) {
-			error("Service record didn't contain the HSP UUID");
-			goto failed_not_supported;
-		}
-		hs->hsp_handle = record->handle;
-	}
-
-	headset_set_channel(hs, record, uuid.value.uuid16);
-
-	if (hs->rfcomm_ch == -1) {
+	if (headset_set_channel(hs, record, p->svclass) < 0) {
 		error("Unable to extract RFCOMM channel from service record");
 		goto failed_not_supported;
 	}
 
+	/* Set svclass to 0 so we can easily check that SDP is no-longer
+	 * going on (to know if bt_cancel_discovery needs to be called) */
+	p->svclass = 0;
+
 	err = rfcomm_connect(dev, NULL, NULL, NULL);
 	if (err < 0) {
 		error("Unable to connect: %s (%d)", strerror(-err), -err);
@@ -1430,16 +1448,16 @@ static void get_record_cb(sdp_list_t *recs, int err, gpointer user_data)
 		goto failed;
 	}
 
-	sdp_list_free(classes, free);
-
 	return;
 
 failed_not_supported:
+	if (p->svclass == HANDSFREE_SVCLASS_ID &&
+			get_records(dev, NULL, NULL, NULL) == 0)
+		return;
 	if (p->msg)
 		error_not_supported(dev->conn, p->msg);
 failed:
-	if (classes)
-		sdp_list_free(classes, free);
+	p->svclass = 0;
 	pending_connect_finalize(dev);
 	headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
 }
@@ -1448,15 +1466,34 @@ static int get_records(struct audio_device *device, headset_stream_cb_t cb,
 			void *user_data, unsigned int *cb_id)
 {
 	struct headset *hs = device->headset;
+	uint16_t svclass;
 	uuid_t uuid;
+	int err;
 
-	sdp_uuid16_create(&uuid, hs->search_hfp ? HANDSFREE_SVCLASS_ID :
-				HEADSET_SVCLASS_ID);
+	if (hs->pending && hs->pending->svclass == HANDSFREE_SVCLASS_ID)
+		svclass = HEADSET_SVCLASS_ID;
+	else
+		svclass = hs->search_hfp ? HANDSFREE_SVCLASS_ID :
+							HEADSET_SVCLASS_ID;
 
-	headset_set_state(device, HEADSET_STATE_CONNECT_IN_PROGRESS);
+	sdp_uuid16_create(&uuid, svclass);
+
+	err = bt_search_service(&device->src, &device->dst, &uuid,
+						get_record_cb, device, NULL);
+	if (err < 0)
+		return err;
+
+	if (hs->pending) {
+		hs->pending->svclass = svclass;
+		return 0;
+	}
+
+	headset_set_state(device, HEADSET_STATE_CONNECTING);
 
 	pending_connect_init(hs, HEADSET_STATE_CONNECTED);
 
+	hs->pending->svclass = svclass;
+
 	if (cb) {
 		unsigned int id;
 		id = connect_cb_new(hs, HEADSET_STATE_CONNECTED,
@@ -1465,8 +1502,7 @@ static int get_records(struct audio_device *device, headset_stream_cb_t cb,
 			*cb_id = id;
 	}
 
-	return bt_search_service(&device->src, &device->dst, &uuid,
-			get_record_cb, device, NULL);
+	return 0;
 }
 
 static int rfcomm_connect(struct audio_device *dev, headset_stream_cb_t cb,
@@ -1475,7 +1511,9 @@ static int rfcomm_connect(struct audio_device *dev, headset_stream_cb_t cb,
 	struct headset *hs = dev->headset;
 	char address[18];
 	GError *err = NULL;
-	GIOChannel *io;
+
+	if (!manager_allow_headset_connection(dev))
+		return -ECONNREFUSED;
 
 	if (hs->rfcomm_ch < 0)
 		return get_records(dev, cb, user_data, cb_id);
@@ -1485,16 +1523,16 @@ static int rfcomm_connect(struct audio_device *dev, headset_stream_cb_t cb,
 	debug("%s: Connecting to %s channel %d", dev->path, address,
 		hs->rfcomm_ch);
 
-	io = bt_io_connect(BT_IO_RFCOMM, headset_connect_cb, dev,
-				NULL, &err,
-				BT_IO_OPT_SOURCE_BDADDR, &dev->src,
-				BT_IO_OPT_DEST_BDADDR, &dev->dst,
-				BT_IO_OPT_CHANNEL, hs->rfcomm_ch,
-				BT_IO_OPT_INVALID);
+	hs->tmp_rfcomm = bt_io_connect(BT_IO_RFCOMM, headset_connect_cb, dev,
+					NULL, &err,
+					BT_IO_OPT_SOURCE_BDADDR, &dev->src,
+					BT_IO_OPT_DEST_BDADDR, &dev->dst,
+					BT_IO_OPT_CHANNEL, hs->rfcomm_ch,
+					BT_IO_OPT_INVALID);
 
 	hs->rfcomm_ch = -1;
 
-	if (!io) {
+	if (!hs->tmp_rfcomm) {
 		error("%s", err->message);
 		g_error_free(err);
 		return -EIO;
@@ -1502,9 +1540,7 @@ static int rfcomm_connect(struct audio_device *dev, headset_stream_cb_t cb,
 
 	hs->hfp_active = hs->hfp_handle != 0 ? TRUE : FALSE;
 
-	g_io_channel_unref(io);
-
-	headset_set_state(dev, HEADSET_STATE_CONNECT_IN_PROGRESS);
+	headset_set_state(dev, HEADSET_STATE_CONNECTING);
 
 	pending_connect_init(hs, HEADSET_STATE_CONNECTED);
 
@@ -1571,7 +1607,7 @@ static DBusMessage *hs_disconnect(DBusConnection *conn, DBusMessage *msg,
 						".NotConnected",
 						"Device not Connected");
 
-	headset_set_state(device, HEADSET_STATE_DISCONNECTED);
+	headset_shutdown(device);
 	ba2str(&device->dst, hs_address);
 	info("Disconnected from %s, %s", hs_address, device->path);
 
@@ -1606,10 +1642,10 @@ static DBusMessage *hs_connect(DBusConnection *conn, DBusMessage *msg,
 	struct headset *hs = device->headset;
 	int err;
 
-	if (hs->state == HEADSET_STATE_CONNECT_IN_PROGRESS)
+	if (hs->state == HEADSET_STATE_CONNECTING)
 		return g_dbus_create_error(msg, ERROR_INTERFACE ".InProgress",
 						"Connect in Progress");
-	else if (hs->state > HEADSET_STATE_CONNECT_IN_PROGRESS)
+	else if (hs->state > HEADSET_STATE_CONNECTING)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
 						".AlreadyConnected",
 						"Already Connected");
@@ -1618,14 +1654,13 @@ static DBusMessage *hs_connect(DBusConnection *conn, DBusMessage *msg,
 		return g_dbus_create_error(msg, ERROR_INTERFACE ".NotReady",
 					"Telephony subsystem not ready");
 
-	if (!manager_allow_headset_connection(&device->src))
-		return g_dbus_create_error(msg, ERROR_INTERFACE ".NotAllowed",
-						"Too many connected devices");
-
 	device->auto_connect = FALSE;
 
 	err = rfcomm_connect(device, NULL, NULL, NULL);
-	if (err < 0)
+	if (err == -ECONNREFUSED)
+		return g_dbus_create_error(msg, ERROR_INTERFACE ".NotAllowed",
+						"Too many connected devices");
+	else if (err < 0)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
 						".ConnectAttemptFailed",
 						"Connect Attempt Failed");
@@ -1716,7 +1751,7 @@ static DBusMessage *hs_play(DBusConnection *conn, DBusMessage *msg,
 
 	switch (hs->state) {
 	case HEADSET_STATE_DISCONNECTED:
-	case HEADSET_STATE_CONNECT_IN_PROGRESS:
+	case HEADSET_STATE_CONNECTING:
 		return g_dbus_create_error(msg, ERROR_INTERFACE
 						".NotConnected",
 						"Device not Connected");
@@ -2048,6 +2083,26 @@ void headset_update(struct audio_device *dev, uint16_t svc,
 	}
 }
 
+static int headset_close_rfcomm(struct audio_device *dev)
+{
+	struct headset *hs = dev->headset;
+	GIOChannel *rfcomm = hs->tmp_rfcomm ? hs->tmp_rfcomm : hs->rfcomm;
+
+	if (rfcomm) {
+		g_io_channel_shutdown(rfcomm, TRUE, NULL);
+		g_io_channel_unref(rfcomm);
+		hs->tmp_rfcomm = NULL;
+		hs->rfcomm = NULL;
+	}
+
+	hs->data_start = 0;
+	hs->data_length = 0;
+
+	hs->nrec = TRUE;
+
+	return 0;
+}
+
 static void headset_free(struct audio_device *dev)
 {
 	struct headset *hs = dev->headset;
@@ -2060,15 +2115,9 @@ static void headset_free(struct audio_device *dev)
 	if (hs->dc_id)
 		device_remove_disconnect_watch(dev->btd_dev, hs->dc_id);
 
-	if (hs->sco) {
-		g_io_channel_shutdown(hs->sco, TRUE, NULL);
-		g_io_channel_unref(hs->sco);
-	}
+	close_sco(dev);
 
-	if (hs->rfcomm) {
-		g_io_channel_shutdown(hs->rfcomm, TRUE, NULL);
-		g_io_channel_unref(hs->rfcomm);
-	}
+	headset_close_rfcomm(dev);
 
 	g_free(hs);
 	dev->headset = NULL;
@@ -2081,8 +2130,7 @@ static void path_unregister(void *data)
 
 	if (hs->state > HEADSET_STATE_DISCONNECTED) {
 		debug("Headset unregistered while device was connected!");
-		bt_cancel_discovery(&dev->src, &dev->dst);
-		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
+		headset_shutdown(dev);
 	}
 
 	debug("Unregistered interface %s on path %s",
@@ -2243,7 +2291,7 @@ unsigned int headset_request_stream(struct audio_device *dev,
 		hs->dc_timer = 0;
 	}
 
-	if (hs->state == HEADSET_STATE_CONNECT_IN_PROGRESS ||
+	if (hs->state == HEADSET_STATE_CONNECTING ||
 			hs->state == HEADSET_STATE_PLAY_IN_PROGRESS)
 		return connect_cb_new(hs, HEADSET_STATE_PLAYING, cb, user_data);
 
@@ -2251,10 +2299,8 @@ unsigned int headset_request_stream(struct audio_device *dev,
 		if (rfcomm_connect(dev, cb, user_data, &id) < 0)
 			return 0;
 		hs->auto_dc = TRUE;
-	} else if (hs->sco == NULL) {
-		if (sco_connect(dev, cb, user_data, &id) < 0)
-			return 0;
-	}
+	} else if (sco_connect(dev, cb, user_data, &id) < 0)
+		return 0;
 
 	hs->pending->target_state = HEADSET_STATE_PLAYING;
 
@@ -2274,7 +2320,7 @@ unsigned int headset_config_stream(struct audio_device *dev,
 		hs->dc_timer = 0;
 	}
 
-	if (hs->state == HEADSET_STATE_CONNECT_IN_PROGRESS)
+	if (hs->state == HEADSET_STATE_CONNECTING)
 		return connect_cb_new(hs, HEADSET_STATE_CONNECTED, cb,
 					user_data);
 
@@ -2368,26 +2414,6 @@ int headset_connect_sco(struct audio_device *dev, GIOChannel *io)
 	return 0;
 }
 
-static int headset_close_rfcomm(struct audio_device *dev)
-{
-	struct headset *hs = dev->headset;
-	GIOChannel *rfcomm = hs->tmp_rfcomm ? hs->tmp_rfcomm : hs->rfcomm;
-
-	if (rfcomm) {
-		g_io_channel_shutdown(rfcomm, TRUE, NULL);
-		g_io_channel_unref(rfcomm);
-		hs->tmp_rfcomm = NULL;
-		hs->rfcomm = NULL;
-	}
-
-	hs->data_start = 0;
-	hs->data_length = 0;
-
-	hs->nrec = TRUE;
-
-	return 0;
-}
-
 static void disconnect_cb(struct btd_device *btd_dev, gboolean removal,
 				void *user_data)
 {
@@ -2395,7 +2421,7 @@ static void disconnect_cb(struct btd_device *btd_dev, gboolean removal,
 
 	info("Headset: disconnect %s", device->path);
 
-	headset_set_state(device, HEADSET_STATE_DISCONNECTED);
+	headset_shutdown(device);
 }
 
 void headset_set_state(struct audio_device *dev, headset_state_t state)
@@ -2423,7 +2449,7 @@ void headset_set_state(struct audio_device *dev, headset_state_t state)
 					AUDIO_HEADSET_INTERFACE,
 					"Disconnected",
 					DBUS_TYPE_INVALID);
-		if (hs->state > HEADSET_STATE_CONNECT_IN_PROGRESS)
+		if (hs->state > HEADSET_STATE_CONNECTING)
 			emit_property_changed(dev->conn, dev->path,
 					AUDIO_HEADSET_INTERFACE, "Connected",
 					DBUS_TYPE_BOOLEAN, &value);
@@ -2433,7 +2459,7 @@ void headset_set_state(struct audio_device *dev, headset_state_t state)
 		device_remove_disconnect_watch(dev->btd_dev, hs->dc_id);
 		hs->dc_id = 0;
 		break;
-	case HEADSET_STATE_CONNECT_IN_PROGRESS:
+	case HEADSET_STATE_CONNECTING:
 		emit_property_changed(dev->conn, dev->path,
 					AUDIO_HEADSET_INTERFACE, "State",
 					DBUS_TYPE_STRING, &state_str);
@@ -2445,11 +2471,15 @@ void headset_set_state(struct audio_device *dev, headset_state_t state)
 					AUDIO_HEADSET_INTERFACE, "State",
 					DBUS_TYPE_STRING, &state_str);
 		if (hs->state < state) {
-			value = TRUE;
+			if (ag.features & AG_FEATURE_INBAND_RINGTONE)
+				hs->inband_ring = TRUE;
+			else
+				hs->inband_ring = FALSE;
 			g_dbus_emit_signal(dev->conn, dev->path,
 						AUDIO_HEADSET_INTERFACE,
 						"Connected",
 						DBUS_TYPE_INVALID);
+			value = TRUE;
 			emit_property_changed(dev->conn, dev->path,
 						AUDIO_HEADSET_INTERFACE,
 						"Connected",
@@ -2609,6 +2639,17 @@ gboolean headset_get_sco_hci(struct audio_device *dev)
 	return sco_hci;
 }
 
+void headset_shutdown(struct audio_device *dev)
+{
+	struct pending_connect *p = dev->headset->pending;
+
+	if (p && p->msg)
+		error_connection_attempt_failed(dev->conn, p->msg, ECANCELED);
+
+	pending_connect_finalize(dev);
+	headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
+}
+
 int telephony_event_ind(int index)
 {
 	if (!active_devices)
@@ -2662,15 +2703,15 @@ int telephony_incoming_call_ind(const char *number, int type)
 
 	/* With HSP 1.2 the RING messages should *not* be sent if inband
 	 * ringtone is being used */
-	if (!hs->hfp_active && (ag.features & AG_FEATURE_INBAND_RINGTONE))
+	if (!hs->hfp_active && hs->inband_ring)
 		return 0;
 
 	g_free(ag.number);
 	ag.number = g_strdup(number);
 	ag.number_type = type;
 
-	if (ag.features & AG_FEATURE_INBAND_RINGTONE && hs->hfp_active &&
-			hs->state != HEADSET_STATE_PLAYING) {
+	if (hs->inband_ring && hs->hfp_active &&
+					hs->state != HEADSET_STATE_PLAYING) {
 		hs->pending_ring = TRUE;
 		return 0;
 	}
@@ -2686,8 +2727,13 @@ int telephony_calling_stopped_ind(void)
 {
 	struct audio_device *dev;
 
+	if (ag.ring_timer) {
+		g_source_remove(ag.ring_timer);
+		ag.ring_timer = 0;
+	}
+
 	if (!active_devices)
-		return -ENODEV;
+		return 0;
 
 	/* In case SCO isn't fully up yet */
 	dev = active_devices->data;
@@ -2697,11 +2743,6 @@ int telephony_calling_stopped_ind(void)
 
 	dev->headset->pending_ring = FALSE;
 
-	if (ag.ring_timer) {
-		g_source_remove(ag.ring_timer);
-		ag.ring_timer = 0;
-	}
-
 	return 0;
 }
 
@@ -2729,7 +2770,7 @@ int telephony_list_current_call_ind(int idx, int dir, int status, int mode,
 	if (!active_devices)
 		return -ENODEV;
 
-	if (number)
+	if (number && strlen(number) > 0)
 		send_foreach_headset(active_devices, hfp_cmp,
 				"\r\n+CLCC: %d,%d,%d,%d,%d,\"%s\",%d\r\n",
 				idx, dir, status, mode, mprty, number, type);
@@ -2769,7 +2810,8 @@ int telephony_call_waiting_ind(const char *number, int type)
 	if (!active_devices)
 		return -ENODEV;
 
-	send_foreach_headset(active_devices, cwa_cmp, "\r\n+CCWA: %s,%d\r\n",
+	send_foreach_headset(active_devices, cwa_cmp,
+				"\r\n+CCWA: \"%s\",%d\r\n",
 				number, type);
 
 	return 0;
