@@ -485,7 +485,7 @@ static void ag_confirm(GIOChannel *chan, gpointer data)
 		goto drop;
 	}
 
-	headset_set_state(device, HEADSET_STATE_CONNECT_IN_PROGRESS);
+	headset_set_state(device, HEADSET_STATE_CONNECTING);
 
 	perr = audio_device_request_authorization(device, server_uuid,
 						headset_auth_cb, device);
@@ -1030,7 +1030,8 @@ static struct btd_adapter_driver avrcp_server_driver = {
 	.remove	= avrcp_server_remove,
 };
 
-int audio_manager_init(DBusConnection *conn, GKeyFile *conf)
+int audio_manager_init(DBusConnection *conn, GKeyFile *conf,
+							gboolean *enable_sco)
 {
 	char **list;
 	int i;
@@ -1116,6 +1117,8 @@ proceed:
 
 	btd_register_device_driver(&audio_driver);
 
+	*enable_sco = (enabled.gateway || enabled.headset);
+
 	return 0;
 }
 
